_MAJOR = "1"
_MINOR = "0"
_PATCH = "1"
# This is mainly for pre-releases which have the suffix "rc[0-9]+".
_SUFFIX = ""

VERSION_SHORT = "{0}.{1}".format(_MAJOR, _MINOR)
VERSION = "{0}.{1}.{2}{3}".format(_MAJOR, _MINOR, _PATCH, _SUFFIX)
